import torch
from torch.autograd import Variable
import torch.nn.functional as F
from .performance import calculate_performance
from sklearn.metrics import classification_report, confusion_matrix, accuracy_score

def test(dataloader, net, criterion, optimizer, opt):
    # net.train()
    test_loss = 0
    correct = 0
    net.eval()
    # print("Len test data : " + str(len(dataloader.dataset)))

    all_targets = []
    all_predicted = []
    for i, (adj_matrices, target) in enumerate(dataloader, 0):
     
        net.zero_grad()
        # optimizer.zero_grad()
        # print("------------------")
        left_adj_matrix = adj_matrices[0]
        right_adj_matrix = adj_matrices[1]
    
        left_init_input = torch.zeros(len(left_adj_matrix), opt.n_node, opt.state_dim).double()
        right_init_input = torch.zeros(len(right_adj_matrix), opt.n_node, opt.state_dim).double()

        if opt.cuda:
            # print("Using cuda for training.......")
            left_init_input = left_init_input.cuda()
            right_init_input = right_init_input.cuda()
            left_adj_matrix = left_adj_matrix.cuda()
            right_adj_matrix = right_adj_matrix.cuda()
            target = target.cuda()

        left_init_input = Variable(left_init_input)
        right_init_input = Variable(right_init_input)

        left_adj_matrix = Variable(left_adj_matrix)
        right_adj_matrix = Variable(right_adj_matrix)

        target = Variable(target)
        
        if opt.loss == 1:
            left_output, right_output = net(left_init_input, left_adj_matrix, right_init_input, right_adj_matrix)
            loss = criterion(left_output,right_output, target) 

            euclidean_distance = F.pairwise_distance(left_output, right_output)   
            
            pred = output.data.max(1, keepdim=True)[1]
            correct += pred.eq(target.data.view_as(pred)).cpu().sum()

    
        else:
            output = net(left_init_input, left_adj_matrix, right_init_input, right_adj_matrix)
            loss = criterion(output, target)

            pred = output.data.max(1, keepdim=True)[1]
            # print(pred)
            # print(pred.data.view_as(target))
           
            correct += pred.eq(target.data.view_as(pred)).cpu().sum()

            all_predicted.extend(pred.data.view_as(target).cpu().numpy())
            all_targets.extend(target.cpu().numpy())

        test_loss += loss.item()
    
    TP, FP, TN, FN = calculate_performance(all_targets, all_predicted)

    precision = 0
    if TP + FP > 0:
       precision = TP / (TP + FP)
    recall = 0
    if TP + FN > 0:
       recall = TP / (TP + FN)
    
    test_loss /= len(dataloader.dataset)

    print('Test set: Average loss: {:.4f}, Accuracy: {}/{} ({:.0f}%), Precision: ({:.0f}%), Recall: ({:.0f}%)'.format(
        test_loss, correct, len(dataloader.dataset), 100. * correct / len(dataloader.dataset), 100. * precision, 100. * recall))



#!/usr/bin/env python3
import argparse
import os
import os.path
import pickle
from collections import OrderedDict

# using mapping to align the lang2 with lang1
def aligning(source, target, mapping):
    print("%s %s" % (source, target))
    with open(source, 'r') as s:
      with open(target, 'w') as t:
          for line in s.readlines():
             numbers = line.rstrip().split()
             if len(numbers)>2 and numbers[0]!='?':
                 line_t = "%s %s %s\n" % (mapping[numbers[0]], numbers[1], mapping[numbers[2]])
             else:
                 line_t = line
             t.write(line_t)
          s.close()
          t.close()

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--lang1', default='cpp', help='language 1')
    parser.add_argument('--lang2', default='java', help='language 2')
    parser.add_argument('--output', default='maps', help='output')
    parser.add_argument('data', nargs='+', help='data sets')
    opt = parser.parse_args()
    print(opt)
    maps = {}
    maps1 = {}
    maps1_filename = 'maps.%s.pkl' % opt.lang1
    if os.path.exists(maps1_filename):
       with open(maps1_filename, 'rb') as f:
            maps1 = pickle.load(f)
    OrderedDict(sorted(maps1.items(), key=lambda t: t[0]))
    maps2_filename = 'maps.%s.pkl' % opt.lang2
    if os.path.exists(maps2_filename):
       with open(maps2_filename, 'rb') as f:
            maps2 = pickle.load(f)
    mapping = {}
    for key, value in maps2.items():
        if key in maps1.keys():
            mapping[value] = maps1[key]
        else:
            maps1[key] = len(mapping.keys())
            mapping[value] = maps1[key]
    #for key, value in mapping.items():
    #    print("%s=>%s" % (key, value))
    from_folder=opt.data[0]
    to_folder=opt.data[1]

    for folder in ['train', 'test']:
       for f in os.listdir(os.path.join(from_folder, folder)):
           if not os.path.exists(os.path.join(to_folder, folder)):
              os.makedirs(os.path.join(to_folder, folder))
           aligning(os.path.join(from_folder, folder, f), os.path.join(to_folder, folder, f), mapping)

if __name__ == "__main__":
    main()



import flatast
import os
import sys
import pickle
sys.path.insert(0, os.path.dirname(flatast.__file__))
from fast_pb2 import Data as PBData
from fast_.Element import *
from fast_.Data import *
from google.protobuf import text_format
import flatbuffers
#import pyarrow

# write data as flatbuffers Data
def _serialize_Data(data):
    return data

# read data as flatbuffers Data
def _deserialize_Data(data):
    buf = bytearray(data)
    return Data.GetRootAsData(buf, 0)

def main():
    #context = pyarrow.SerializationContext()
    #context.register_type(Data, 'Data',
    #    custom_serializer=_serialize_Data,
    #    custom_deserializer=_deserialize_Data)

    if sys.argv[1].endswith('.pb') and (sys.argv[2].endswith('.pickle') or sys.argv[2].endswith('.pkl')):
        data = PBData()
        with open(sys.argv[1], 'rb') as f:
            data.ParseFromString(f.read())
            f.close()
        with open(sys.argv[2], 'wb') as f:
            pickle.dump(data, f);
    #elif sys.argv[1].endswith('.fbs') and sys.argv[2].endswith('.pa'):
    #    with open(sys.argv[1], 'rb') as f:
    #        data = f.read()
    #        buf = bytearray(data)
    #        data = Data.GetRootAsData(buf, 0)
    #        f.close()
    #    with open(sys.argv[2], 'wb') as f:
    #        buf = context.serialize(data).to_buffer()
    #        f.write(buf)
    #        f.close()
    elif sys.argv[2].endswith('.pb') and (sys.argv[1].endswith('.pickle') or sys.argv[1].endswith('.pkl')):
        with open(sys.argv[1], 'rb') as f:
            data = pickle.load(f);
        with open(sys.argv[2], 'wb') as f:
            serializedMessage = data.SerializeToString()
            f.write(serializedMessage)
            f.close()
    #elif sys.argv[2].endswith('.fbs') and sys.argv[1].endswith('.pa'):
    #    with open(sys.argv[1], 'rb') as f:
    #        buf = f.read()
    #        data = context.deserialize(buf)
    #        f.close()
    #    with open(sys.argv[2], 'wb') as f:
    #        builder = flatbuffers.Builder(0)
    #        builder.Finish(data)
            buf = builder.Output()
            f.write(buf)
            f.close()
    else:
        print("Please choose either .pb, .pkl, or .pickle as file extensions")

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import ggnn
import argparse
import random
import torch
import torch.nn as nn
import torch.optim as optim
import sqlite3
import utils.data
from utils.model import GGNN
from utils.model import ClassPrediction
from utils.train_ggnn import train
from utils.test_ggnn import test
from utils.data import dataset
# from utils.data.dataset import MonoLanguageProgramData
from utils.data.dataset import ProgramData
from utils.data.dataloader import bAbIDataloader
from tensorboardX import SummaryWriter
import os
import os.path
import json
import pickle
import time
import sys
import tempfile
import hashlib
import flatbuffers
from tqdm import tqdm
import fast_
import fast_.Graph_
import fast_.Data_
import fast_.Data_.Anonymous3
from fast_ import Data, Graph, Element, Bugs, Log, Delta
from fast_.Data_ import Anonymous3
from fast_.Data import *
from fast_.Graph import *
from fast_.Element import *
from fast_.Bugs import *
from fast_.Log import *
from fast_.Delta import *
from fast_.Data_.Anonymous3 import *
from fast_.Graph import *
from fast_.Graph_.Unit import *
from fast_.Graph_.ContextGraph import *
from fast_.Graph_.ContextEdges import *
from fast_.Graph_.Edge import *
from fast_.Graph_.SymbolCandidate import *
from fast_.Graph_.NodeLabel import *
from fast_.Graph_.NodeType import *
from flatast.json_format import MessageToJson
import flatast.fast_pb2
import traceback
import pdb
import tqdm


parser = argparse.ArgumentParser()
parser.add_argument('--db', default="gitlog.db", help='the data sources')
parser.add_argument('--manualSeed', type=int, help='manual seed')
parser.add_argument('--label', default=None, help='assign the datasets to this class')
parser.add_argument('--n_classes', type=int, default=0, help='number of classes')
parser.add_argument('argv', nargs="*", help='add all the files if n_classes = 0')

# TODO: keep relevant options

# For feature engineering
parser.add_argument('--maps', action='store_true', default=True, help='maps node type to a consequetive number')
parser.add_argument('--map_folder', default=".", help='where to store the maps file')
parser.add_argument('--edgemaps', action='store_true', default=False, help='maps node type to a consequetive number')
############ Do not remove duplicated edges, which is found irrelevant to the performance
parser.add_argument('--dup', action='store_true', default=False, help='keep duplicated edges of the nodetypes')
############ Reinitialise the maps per class of instances. It leads to inappropriate performance for the classification
parser.add_argument('--localmaps', action='store_true', default=False, help='use local maps instead of global one')
## Use the occurrence of type instead of label to encode nodes
parser.add_argument('--occurrence', type=bool, default=True, help='use the <nodetype, occurrence> representation of nodes')
parser.add_argument('--mod', type=int, default=2, help='use the <nodetype, occurrence % mod> representation of nodes')
## Ignore the POSITION and COMMENT node types as noises to the input
parser.add_argument('--noposition', type=bool, default=True, help='ignore POSITION and COMMENT node types')
## Derive the last lexical use edges from adjacent occurrences of same node encodings
parser.add_argument('--lastuse', type=bool, default=False, help='add lastLexicalUse edges "3" between adjacent occurrence nodes  of the same node labels')
## Consider only the edge types of Child and NextToken
parser.add_argument('--syntaxonly', type=bool, default=False, help='output only syntactical edges, i.e. type "1" and "2"')
## Do not distinguish edge types
parser.add_argument('--noedgetype', type=bool, default=False, help='ignore edge types by replacing all edge types with "1"')
## Add label to the encoding of nodes
parser.add_argument('--labelling', type=bool, default=False, help='use the <nodetype, nodelabel> representation of nodes')
## Generate bidirectional edges
parser.add_argument('--bidirect', type=bool, default=False, help='make edges bidirectional')
## Add extra semantic edges as if they are syntactical
parser.add_argument('--mixing', type=bool, default=False, help='make semantic edges syntactical to allow for propagation')

# For learning
parser.add_argument('--workers', type=int, help='number of data loading workers', default=2)
parser.add_argument('--train_batch_size', type=int, default=32, help='input batch size')
parser.add_argument('--test_batch_size', type=int, default=32, help='input batch size')
parser.add_argument('--state_dim', type=int, default=5, help='GGNN hidden state size')
parser.add_argument('--n_steps', type=int, default=10, help='propogation steps number of GGNN')
parser.add_argument('--niter', type=int, default=150, help='number of epochs to train for')
parser.add_argument('--lr', type=float, default=0.01, help='learning rate')
parser.add_argument('--cuda', action='store_true', help='enables cuda')
parser.add_argument('--verbal', type=bool, default=True, help='print training info or not')
# parser.add_argument('--directory', default="program_data/cpp_babi_format_Sep-29-2018-0000006", help='program data')
parser.add_argument('--model_path', default="model/model.ckpt", help='path to save the model')
parser.add_argument('--n_hidden', type=int, default=50, help='number of hidden layers')
parser.add_argument('--size_vocabulary', type=int, default=59, help='maximum number of node types')
parser.add_argument('--is_training_ggnn', type=bool, default=True, help='Training GGNN or BiGGNN')
parser.add_argument('--log_path', default="" ,help='log path for tensorboard')
parser.add_argument('--epoch', type=int, default=0, help='epoch to test')

opt = parser.parse_args()
# print(opt)
if opt.manualSeed is None:
    opt.manualSeed = random.randint(1, 10000)
    # print("Random Seed: ", opt.manualSeed)
random.seed(opt.manualSeed)


def fbsEdges(builder, edges, type):
    typed_edges = edges[type]
    fbs_edges = []
    for e in typed_edges:
        EdgeStart(builder)
        EdgeAddNode1(builder, e[0])
        EdgeAddNode2(builder, e[1])
        fbs_edges.append(EdgeEnd(builder))
    N = len(fbs_edges)
    if type == "NextToken":
        ContextEdgesStartNextTokenVector(builder, N)
    elif type == "Child":
        ContextEdgesStartChildVector(builder, N)
    elif type == "LastWrite":
        ContextEdgesStartLastWriteVector(builder, N)
    elif type == "ReturnsTo":
        ContextEdgesStartReturnsToVector(builder, N)
    elif type == "LastUse":
        ContextEdgesStartLastUseVector(builder, N)
    elif type == "LastLexicalUse":
        ContextEdgesStartLastLexicalUseVector(builder, N)
    elif type == "ComputesFrom":
        ContextEdgesStartComputesFromVector(builder, N)
    for i in reversed(range(0, N)):
       builder.PrependUOffsetTRelative(fbs_edges[i])
    fbs_nodeTypes = builder.EndVector(N)
    return fbs_nodeTypes

def fbsContextEdges(builder, edges):
    #print(edges)
    if 'NextToken' in edges:
        nextToken = fbsEdges(builder, edges, 'NextToken')
    if 'Child' in edges:
        child = fbsEdges(builder, edges, 'Child')
    if 'LastLexicalUse' in edges:
        lastLexicalUse = fbsEdges(builder, edges, 'LastLexicalUse')
    if 'LastUse' in edges:
        lastUse = fbsEdges(builder, edges, 'LastUse')
    if 'LastWrite' in edges:
        lastWrite = fbsEdges(builder, edges, 'LastWrite')
    if 'ReturnsTo' in edges:
        returnsTo = fbsEdges(builder, edges, 'ReturnsTo')
    if 'ComputesFrom' in edges:
        computesFrom = fbsEdges(builder, edges, 'ComputesFrom')
    ContextEdgesStart(builder)
    if 'NextToken' in edges:
        ContextEdgesAddNextToken(builder, nextToken)
    if 'Child' in edges:
        ContextEdgesAddChild(builder, child)
    if 'LastLexicalUse' in edges:
        ContextEdgesAddLastLexicalUse(builder, lastLexicalUse)
    if 'LastUse' in edges:
        ContextEdgesAddLastUse(builder, lastUse)
    if 'LastWrite' in edges:
        ContextEdgesAddLastWrite(builder, lastWrite)
    if 'ReturnsTo' in edges:
         ContextEdgesAddReturnsTo(builder, returnsTo)
    if 'ComputesFrom' in edges:
         ContextEdgesAddComputesFrom(builder, computesFrom)
    return ContextEdgesEnd(builder)

def fbsNodeType(builder, key, value):
    fbs_key = int(key)
    fbs_value = builder.CreateString(value)
    NodeTypeStart(builder)
    NodeTypeAddNode(builder, fbs_key)
    NodeTypeAddType(builder, fbs_value)
    return NodeTypeEnd(builder)

def fbsNodeTypes(builder, node_types):
    nodeTypes = []
    for k in node_types.keys():
        nodeTypes.append(fbsNodeType(builder, k, node_types[k]))
    N = len(nodeTypes)
    ContextGraphStartNodeTypeVector(builder, N)
    for i in reversed(range(0, N)):
       builder.PrependUOffsetTRelative(nodeTypes[i])
    fbs_nodeTypes = builder.EndVector(N)
    return fbs_nodeTypes

def fbsNodeLabel(builder, key, value):
    fbs_key = int(key)
    fbs_value = builder.CreateString(value)
    NodeLabelStart(builder)
    NodeLabelAddNode(builder, fbs_key)
    NodeLabelAddLabel(builder, fbs_value)
    return NodeLabelEnd(builder)

def fbsNodeLabels(builder, node_labels):
    nodeLabels = []
    for k in node_labels.keys():
        nodeLabels.append(fbsNodeLabel(builder, k, node_labels[k]))
    N = len(nodeLabels)
    ContextGraphStartNodeLabelVector(builder, N)
    for i in reversed(range(0, N)):
       builder.PrependUOffsetTRelative(nodeLabels[i])
    fbs_nodeLabels = builder.EndVector(N)
    return fbs_nodeLabels

def fbsGraph(builder, graph):
    fbs_edges = fbsContextEdges(builder, graph["Edges"])
    fbs_nodeLabels = fbsNodeLabels(builder, graph["NodeLabels"])
    fbs_nodeTypes = fbsNodeTypes(builder, graph["NodeTypes"])

    ContextGraphStart(builder)
    ContextGraphAddEdges(builder, fbs_edges)
    ContextGraphAddNodeLabel(builder, fbs_nodeLabels)
    ContextGraphAddNodeType(builder, fbs_nodeTypes)
    fbs_graph = ContextGraphEnd(builder)
    return fbs_graph

def fbsSymbolCandidate(builder, symbolCandidate):
    symbolName = builder.CreateString(symbolCandidate['SymbolName'])
    SymbolCandidateStart(builder)
    SymbolCandidateAddSymbolDummyNode(builder, symbolCandidate["SymbolDummyNode"])
    SymbolCandidateAddSymbolName(builder, symbolName)
    SymbolCandidateAddIsCorrect(builder, symbolCandidate["IsCorrect"])
    symbolCandidates = SymbolCandidateEnd(builder)
    return symbolCandidates

def fbsSymbolCandidates(builder, symbolCandidates):
    candidates = []
    for s in symbolCandidates:
        candidates.append(fbsSymbolCandidate(builder, s))
    N = len(candidates)
    UnitStartSymbolCandidateVector(builder, N)
    for i in reversed(range(0, N)):
       builder.PrependUOffsetTRelative(candidates[i])
    fbs_candidates = builder.EndVector(N)
    return fbs_candidates

def fbsGatedGraph(builder, data):
    units = []
    for g in data:
        fbs_filename = builder.CreateString(g['filename'])
        fbs_slotTokenIdx = g['slotTokenIdx']
        fbs_SlotDummyNode = g['SlotDummyNode']
        fbs_Graph = fbsGraph(builder, g['ContextGraph'])
        fbs_SymbolCandidate = fbsSymbolCandidates(builder, g['SymbolCandidates'])
        UnitStart(builder)
        UnitAddFilename(builder, fbs_filename)
        UnitAddSlotTokenIdx(builder, fbs_slotTokenIdx)
        UnitAddGraph(builder, fbs_Graph)
        UnitAddSlotDummyNode(builder, fbs_SlotDummyNode)
        UnitAddSymbolCandidate(builder, fbs_SymbolCandidate)
        unit = UnitEnd(builder)
        units.append(unit)
    N = len(units)
    GraphStartUnitVector(builder, N)
    for i in tqdm(reversed(range(0, N))):
        builder.PrependUOffsetTRelative(units[i])
    fbs_unit = builder.EndVector(N)
    GraphStart(builder)
    GraphAddUnit(builder, fbs_unit)
    graph = GraphEnd(builder)
    DataStart(builder)
    DataAddRecordType(builder, graph)
    data = DataEnd(builder)
    return data

def jdefault(o):
    if o is None:
        return ' '
    if isinstance(o, set):
        return list(o)
    elif isinstance(o, Data):
        return jdefault(o.RecordType())
    elif isinstance(o, Anonymous3):
        return jdefault(o.Graph())
    elif isinstance(o, Graph):
        obj = []
        for i in range(0, o.UnitLength()):
            obj.append(jdefault(o.Unit(i)))
        return obj
    elif isinstance(o, Unit):
        obj = {}
        obj['filename'] = o.Filename()
        obj['slotTokenIdx'] = o.SlotTokenIdx()
        obj['ContextGraph'] = o.Graph()
        obj['SlotDummyNode'] = o.SlotDummyNode()
        vec = []
        for i in range(0, o.SymbolCandidateLength()):
            vec.append(jdefault(o.SymbolCandidate(i)))
        obj['SymbolCandidates'] = vec
        return obj
    elif isinstance(o, SymbolCandidate):
        obj = {}
        obj['SymbolDummyNode'] = o.SymbolDummyNode()
        obj['SymbolName'] = o.SymbolName()
        obj['IsCorrect'] = o.IsCorrect()
        return obj
    elif isinstance(o, ContextGraph):
        obj = {}
        obj['Edges'] = o.Edges()
        array = {}
        for i in range(0, o.NodeLabelLength()):
            nl = o.NodeLabel(i)
            array[nl.Node()] = nl.Label()
        obj['NodeLabels'] = array
        array = {}
        for i in range(0, o.NodeTypeLength()):
            nl = o.NodeType(i)
            array[nl.Node()] = nl.Type()
        obj['NodeTypes'] = array
        return obj
    elif isinstance(o, ContextEdges):
        obj = {}
        vec = []
        for i in range(0, o.NextTokenLength()):
            vec.append(jdefault(o.NextToken(i)))
        obj['NextToken'] = vec
        vec = []
        for i in range(0, o.ChildLength()):
            vec.append(jdefault(o.Child(i)))
        obj['Child'] = vec
        vec = []
        for i in range(0, o.LastLexicalUseLength()):
            vec.append(jdefault(o.LastLexicalUse(i)))
        obj['LastLexicalUse'] = vec
        vec = []
        for i in range(0, o.LastUseLength()):
            vec.append(jdefault(o.LastUse(i)))
        obj['LastUse'] = vec
        vec = []
        for i in range(0, o.LastWriteLength()):
            vec.append(jdefault(o.LastWrite(i)))
        obj['LastWrite'] = vec
        vec = []
        for i in range(0, o.ReturnsToLength()):
            vec.append(jdefault(o.ReturnsTo(i)))
        obj['ReturnsTo'] = vec
        vec = []
        for i in range(0, o.ComputesFromLength()):
            vec.append(jdefault(o.ComputesFrom(i)))
        obj['ComputesFrom'] = vec
        return obj
    elif isinstance(o, Edge):
        obj = [o.Node1(), o.Node2()]
        return obj
    elif isinstance(o, bytes):
        return str(o)
    elif isinstance(o, str):
        return o
    return o.__dict__

def add_edge_to_json(opt, edges, out, uniq_edges, src, edgetype, tgt):
    if src != tgt:
       e="%s %d %s\n" % (src, edgetype, tgt)
       if opt.dup:
          edges.append(e)
       else:
          uniq_edges.add(e)

def ggnn2json(data, map_folder='.'):
    graph = data.RecordType().Graph()
    if opt.maps or opt.edgemaps:
        maps = {}
    algorithms = []
    node2type = True
    train_graph = []
    test_graph = []
    for i in range(0, graph.UnitLength()):
        edges = []
        unit = graph.Unit(i)
        p = unit.Filename()
        t = os.path.dirname(p)
        t = os.path.dirname(t)
        if t in algorithms:
            algo = algorithms.index(t)
        else:
            algo = len(algorithms)
            algorithms.append(t)
        if (opt.maps or opt.edgemaps) and not opt.localmaps:
            input_basename, input_extension = os.path.splitext(p)
            maps_filename = "%s/maps%s.pkl" % (map_folder, input_extension.decode('ASCII'))
            if os.path.exists(maps_filename):
              with open(maps_filename, 'rb') as f:
                 maps = pickle.load(f)
        g = unit.Graph()
        edges = g.Edges()
        NT = g.NodeTypeLength()
        dict = {}
        if opt.occurrence:
            occurrence = {}
        uniq_edges = {}
        labels = {}
        for j in range(0, g.NodeLabelLength()):
            nl = g.NodeLabel(j)
            t = nl.Label()
            j1 = str(j + 1)
            labels[j1] = t
        ids = {}
        types = {}
        for j in range(0, g.NodeTypeLength()):
            nl = g.NodeType(j)
            t = str(nl.Type().decode('ASCII')) 
            j1 = str(j + 1)
            if t != 'POSITION' and t != 'COMMENT' and t != '271' and t != '6':
               types[j1] = t
               to = t
               if opt.labelling:
                  l = labels[j1]
                  if l != b'':
                     to = "%s:%s" % (t, l)
               elif opt.occurrence:
                  if not t in occurrence.keys():
                     occurrence[t] = 1 
                  else:
                     occurrence[t] = occurrence[t] + 1
                  to = "%s:%d" % (t, occurrence[t] % opt.mod)
               dict[j1] = to
               if opt.maps or opt.edgemaps:
                  if not to in maps:
                     maps[to] = str(1 + len(maps))
               if t == 'NAME':
                  ids[j1] = dict[j1] #labels[j1]
               if opt.edgemaps:
                  add_edge_to_json(opt, edges, uniq_edges, j1, 7 + int(maps[to]), j1)
            else:
               dict[j1] = 0
        for edgetype in range(1, 8):
            # out.write("====== %d\n" % edgetype)
            if edgetype == 1:
                n = edges.ChildLength()
            elif edgetype == 2:
                n = edges.NextTokenLength()
            elif edgetype == 3:
                n = edges.LastLexicalUseLength()
            elif edgetype == 4:
                n = edges.LastUseLength()
            elif edgetype == 5:
                n = edges.LastWriteLength()
            elif edgetype == 6:
                n = edges.ReturnsToLength()
            elif edgetype == 7:
                n = edges.ComputesFromLength()
            for j in range(0, n):
                if edgetype == 1:
                    e = edges.Child(j)
                elif edgetype == 2:
                    e = edges.NextToken(j)
                elif edgetype == 3:
                    e = edges.LastLexicalUse(j)
                elif edgetype == 4:
                    e = edges.LastUse(j)
                elif edgetype == 5:
                    e = edges.LastWrite(j)
                elif edgetype == 6:
                    e = edges.ReturnsTo(j)
                elif edgetype == 7:
                    e = edges.ComputesFrom(j)
                if opt.maps and not opt.edgemaps:
                   src = dict[str(e.Node1())]
                   tgt = dict[str(e.Node2())] 
                else:
                   src = str(e.Node1())
                   tgt = str(e.Node2())
                if src != 0 and tgt != 0:
                    if opt.maps:
                        s1 = maps[src]
                    else:
                        s1 = src
                    if opt.noedgetype:
                        edgetype = 1
                    else:
                        s2 = str(edgetype)
                    if opt.maps:
                        s3 = maps[tgt]
                    else:
                        s3 = tgt
                    if (s2 == '1' or s2 == '2') or not opt.syntaxonly:
                       add_edge_to_json(opt, edges, uniq_edges, s1, edgetype, s3)
                       if opt.bidirect:
                          add_edge_to_json(opt, edges, uniq_edges, s3, edgetype, s1)
                       if opt.mixing:
                          add_edge_to_json(opt, edges, uniq_edges, s1, 1, s3)
        if opt.lastuse:
            lastindex = {}
            lastuses = {}
            for j in range(0, g.NodeLabelLength()):
                nl = g.NodeLabel(j)
                t = nl.Label()
                j1 = str(j + 1)
                if t != b'' and t != b'int':
                   if t in lastindex.keys():
                      lastuses[j1] = lastindex[t]
                   lastindex[t] = j1
            for k, v in lastuses.items():
               if opt.maps:
                  t1 = dict[k]
                  t2 = dict[v]
               else:
                  t1 = str(k)
                  t2 = str(v)
               if t1 != '0' and t2 != '0' and t1 != 0 and t2 != 0:
                  if opt.maps:
                     add_edge_to_json(opt, edges, uniq_edges, maps[t2], 3, maps[t1])
                  else:
                     add_edge_to_json(opt, edges, uniq_edges, t2, 3, t1)
        out = {}
        out["graph"] = []
        for e in uniq_edges.keys():
            [src, edgetype, tgt] = e.split()
            out["graph"].append([int(src), int(edgetype), int(tgt)])
        out["target"] = []
        out["target"].append([i])
        out["node_features"] = []
        if i % 5 != 0:
            train_graph.append(out)
        else:
            test_graph.append(out)
        if (opt.maps or opt.edgemaps) and not opt.localmaps:
            # Don't assume the files in the same dataset are of the same language
            with open(maps_filename, 'wb') as f:
                pickle.dump(maps, f, 2)
    return [train_graph, test_graph]

def add_edge(opt, uniq_edges, src, edgetype, tgt):
    if src != tgt:
        e="%s %d %s" % (src, edgetype, tgt)
        # print(e)
        uniq_edges.add(e)

maps = {}

def ggnn2str(opt, unit, uniqe_edges, map_folder=".", algorithms=[], maps = {}):
    p = unit.Filename()
    t = os.path.dirname(p)
    t = os.path.dirname(t)
    uniq_edges = set()
    if t in algorithms:
        i = algorithms.index(t)
    else:
        i = len(algorithms)
        algorithms.append(t)
    if (opt.maps or opt.edgemaps) and not opt.localmaps:
        _, input_extension = os.path.splitext(p)
        maps_filename = "%s/maps%s.pkl" % (map_folder, input_extension.decode('ASCII'))
        if os.path.exists(maps_filename):
            with open(maps_filename, 'rb') as f:
                maps = pickle.load(f)
    g = unit.Graph()
    edges = g.Edges()

    dict = {}
    if opt.occurrence:
        occurrence = {}
    labels = {}
    for j in range(0, g.NodeLabelLength()):
        nl = g.NodeLabel(j)
        t = nl.Label()
        j1 = str(j + 1)
        labels[j1] = t
    ids = {}
    types = {}
    for j in range(0, g.NodeTypeLength()):
        nl = g.NodeType(j)
        t = str(nl.Type().decode('ASCII')) 
        j1 = str(j + 1)
        if t != 'POSITION' and t != 'COMMENT' and t != '271' and t != '6':
            types[j1] = t
            to = t
            if opt.labelling:
                l = labels[j1]
                if l != b'':
                    to = "%s:%s" % (t, l)
            elif opt.occurrence:
                if not t in occurrence.keys():
                    occurrence[t] = 1 
                else:
                    occurrence[t] = occurrence[t] + 1
                to = "%s:%d" % (t, occurrence[t] % opt.mod)
            dict[j1] = to
            if opt.maps or opt.edgemaps:
                if not to in maps:
                    maps[to] = str(1 + len(maps))
            if t == 'NAME':
                ids[j1] = dict[j1] #labels[j1]
            if opt.edgemaps:
                add_edge(opt, uniq_edges, j1, 7 + int(maps[to]), j1)
        else:
            dict[j1] = 0
    for edgetype in range(1, 8):
        if edgetype == 1:
            n = edges.ChildLength()
        elif edgetype == 2:
            n = edges.NextTokenLength()
        elif edgetype == 3:
            n = edges.LastLexicalUseLength()
        elif edgetype == 4:
            n = edges.LastUseLength()
        elif edgetype == 5:
            n = edges.LastWriteLength()
        elif edgetype == 6:
            n = edges.ReturnsToLength()
        elif edgetype == 7:
            n = edges.ComputesFromLength()
        for j in range(0, n):
            if edgetype == 1:
                e = edges.Child(j)
            elif edgetype == 2:
                e = edges.NextToken(j)
            elif edgetype == 3:
                e = edges.LastLexicalUse(j)
            elif edgetype == 4:
                e = edges.LastUse(j)
            elif edgetype == 5:
                e = edges.LastWrite(j)
            elif edgetype == 6:
                e = edges.ReturnsTo(j)
            elif edgetype == 7:
                e = edges.ComputesFrom(j)
            if opt.maps and not opt.edgemaps:
                src = dict[str(e.Node1())]
                tgt = dict[str(e.Node2())] 
            else:
                src = str(e.Node1())
                tgt = str(e.Node2())
            if src != 0 and tgt != 0:
                if opt.maps:
                    s1 = maps[src]
                else:
                    s1 = src
                if opt.noedgetype:
                    edgetype = 1
                else:
                    s2 = str(edgetype)
                if opt.maps:
                    s3 = maps[tgt]
                else:
                    s3 = tgt
                if (s2 == '1' or s2 == '2') or not opt.syntaxonly:
                    add_edge(opt, uniq_edges, s1, edgetype, s3)
                    if opt.bidirect:
                        add_edge(opt, uniq_edges, s3, edgetype, s1)
                    if opt.mixing:
                        add_edge(opt, uniq_edges, s1, 1, s3)
    if opt.lastuse:
        lastindex = {}
        lastuses = {}
        for j in range(0, g.NodeLabelLength()):
            nl = g.NodeLabel(j)
            t = nl.Label()
            j1 = str(j + 1)
            if t != b'' and t != b'int':
                if t in lastindex.keys():
                    lastuses[j1] = lastindex[t]
                lastindex[t] = j1
        for k, v in lastuses.items():
            if opt.maps:
                t1 = dict[k]
                t2 = dict[v]
            else:
                t1 = str(k)
                t2 = str(v)
            if t1 != '0' and t2 != '0' and t1 != 0 and t2 != 0:
                if opt.maps:
                    add_edge(opt, uniq_edges, maps[t2], 3, maps[t1])
                else:
                    add_edge(opt, uniq_edges, t2, 3, t1)
    return [maps_filename, uniq_edges]

def ggnn2txt(data, train, test, map_folder='.'):
    graph = data.RecordType().Graph()
    if opt.maps or opt.edgemaps:
        maps = {}
    algorithms = []
    for i in range(0, graph.UnitLength()):
        if (i % 5 != 0 and train != test):
            out = train
        else:
            out = test
        uniq_edges = {}
        unit = graph.unit(i)
        [maps_filename, _] = ggnn.ggnn2str(opt, unit, uniq_edges, map_folder, algorithms, maps)
        for e in uniq_edges.keys():
            out.write(e)
        out.write("? %d %s\n\n" % (i+1, p.decode('ASCII')))
        if (opt.maps or opt.edgemaps) and not opt.localmaps:
            # Don't assume the files in the same dataset are of the same language
            with open(maps_filename, 'wb') as f:
                pickle.dump(maps, f, 2)

def get_descendants(child, node):
    descendants = [node]
    for i in range(0, len(child)):
        edge = child[i]
        if edge[0] == int(node):
            descendants.extend(get_descendants(child, str(edge[1])))
    return descendants

def get_subgraph(graph, subnodes):
    idx = {}
    for i in range(0, len(subnodes)):
        idx[subnodes[i]] = i + 1
    subgraph = {}
    edges = graph["Edges"]
    subedges = {}
    child = edges["Child"]
    subchild = []
    nexttoken = edges["NextToken"]
    subnexttoken = []
    lastlexicaluse = edges["LastLexicalUse"]
    sublastlexicaluse = []
    lastuse = edges["LastUse"]
    sublastuse = []
    lastwrite = edges["LastWrite"]
    sublastwrite = []
    returnsto = edges["ReturnsTo"]
    subreturnsto = []
    computesfrom = edges["ComputesFrom"]
    subcomputesfrom = []
    nodetypes = graph["NodeTypes"]
    subnodetypes = {}
    nodelabels = graph["NodeLabels"]
    subnodelabels = {}
    for i in range(0, len(child)):
        edge = child[i]
        if str(edge[0]) in subnodes and str(edge[1]) in subnodes:
            subedge = []
            subedge.append(idx[str(edge[0])])
            subedge.append(idx[str(edge[1])])
            subchild.append(subedge)
    for i in range(0, len(nexttoken)):
        edge = nexttoken[i]
        if str(edge[0]) in subnodes and str(edge[1]) in subnodes:
            subedge = []
            subedge.append(idx[str(edge[0])])
            subedge.append(idx[str(edge[1])])
            subnexttoken.append(subedge)
    for i in range(0, len(lastlexicaluse)):
        edge = lastlexicaluse[i]
        if str(edge[0]) in subnodes and str(edge[1]) in subnodes:
            subedge = []
            subedge.append(idx[str(edge[0])])
            subedge.append(idx[str(edge[1])])
            sublastlexicaluse.append(subedge)
    for i in range(0, len(lastuse)):
        edge = lastuse[i]
        if str(edge[0]) in subnodes and str(edge[1]) in subnodes:
            subedge = []
            subedge.append(idx[str(edge[0])])
            subedge.append(idx[str(edge[1])])
            sublastuse.append(subedge)
    for i in range(0, len(lastwrite)):
        edge = lastwrite[i]
        if str(edge[0]) in subnodes and str(edge[1]) in subnodes:
            subedge = []
            subedge.append(idx[str(edge[0])])
            subedge.append(idx[str(edge[1])])
            sublastwrite.append(subedge)
    for i in range(0, len(returnsto)):
        edge = returnsto[i]
        if str(edge[0]) in subnodes and str(edge[1]) in subnodes:
            subedge = []
            subedge.append(idx[str(edge[0])])
            subedge.append(idx[str(edge[1])])
            subreturnsto.append(subedge)
    for i in range(0, len(computesfrom)):
        edge = computesfrom[i]
        if str(edge[0]) in subnodes and str(edge[1]) in subnodes:
            subedge = []
            subedge.append(idx[str(edge[0])])
            subedge.append(idx[str(edge[1])])
            subcomputesfrom.append(subedge)
    for k, v in nodetypes.items():
        if k in subnodes:
            subnodetypes[str(idx[k])] = v
    for k, v in nodelabels.items():
        if k in subnodes:
            subnodelabels[str(idx[k])] = v
    subedges["Child"] = subchild
    subedges["NextToken"] = subnexttoken
    subedges["LastLexicalUse"] = sublastlexicaluse
    subedges["LastUse"] = sublastuse
    subedges["LastWrite"] = sublastwrite
    subedges["ReturnsTo"] = subreturnsto
    subedges["ComputesFrom"] = subcomputesfrom
    subgraph["Edges"] = subedges
    subgraph["NodeTypes"] = subnodetypes
    subgraph["NodeLabels"] = subnodelabels
    return subgraph

def get_subgraphs(filename, graph):
    data = []
    edges = graph["Edges"]
    child = edges["Child"]
    nodelabels=graph["NodeLabels"]
    for k, v in nodelabels.items():
        subnodes = get_descendants(child, k)
        #
        # Each candidate should have at least 10 nodes
        #
        if len(subnodes) > 10:
            subgraph = get_subgraph(graph, subnodes)
            subdata = {}
            input_basename, input_extension = os.path.splitext(filename)
            subdata["filename"] = '%s-%s%s' % (input_basename.decode('ASCII'), k, input_extension.decode('ASCII'))
            subdata["ContextGraph"] = subgraph
            subdata["slotTokenIdx"] = 0
            subdata["SlotDummyNode"] = 0
            subdata["SymbolCandidates"] = []
            data.append(subdata)
    return data
  
def generate_subgraphs(filename, graph, out):
    with open('/tmp/t.json', 'w') as json_out:
        json.dump(graph, json_out, default=jdefault)
        json_out.close()
    with open('/tmp/t.json', 'r') as json_in:
        graph_data = json.load(json_in)
        json_in.close()
    data = get_subgraphs(filename, graph_data)
    #json.dump(data, json_out)
    builder = flatbuffers.Builder(0)
    fbs_graph = fbsGatedGraph(builder, data)
    builder.Finish(fbs_graph)
    with open('/tmp/tt.fbs', 'wb') as fbs_out:
        buf = builder.Output()
        fbs_out.write(buf)
        fbs_out.close()
    with open('/tmp/tt.fbs', 'rb') as fbs_in:
        buf = fbs_in.read()
        buf = bytearray(buf)
        fbs_graph = Graph.GetRootAsGraph(buf, 0)
        fbs_in.close()
    ggnn2txt(fbs_graph, out, out)

#
# generate a graph for the AST of each node
#
def ggnn2txt_test(graph, test):
    if opt.maps and not opt.localmaps:
        maps = {}
        if os.path.exists('maps.pkl'):
            with open('maps.pkl', 'rb') as f:
                 maps = pickle.load(f)
    algorithms = []
    node2type = True
    for i in trange(0, graph.UnitLength()):
        out = test
        unit = graph.Unit(i)
        p = unit.Filename()
        t = os.path.dirname(p);
        t = os.path.dirname(t);
        if t in algorithms:
            i = algorithms.index(t)
        else:
            i = len(algorithms)
            algorithms.append(t)
        g = unit.Graph()
        edges = g.Edges()
        array = {}
        NT = g.NodeTypeLength()
        dict = {}
        for j in range(0, g.NodeTypeLength()):
            nl = g.NodeType(j)
            j1 = str(j + 1)
            dict[j1] = str(nl.Type())
            if opt.maps:
                if not str(nl.Type()) in maps:
                    maps[str(nl.Type())] = str(1 + len(maps))
        generate_subgraphs(p, g, out)
    if not opt.localmaps:
        with open('maps.pkl', 'wb') as f:
            pickle.dump(maps, f, 2)



def get_temp_filename(ext):
    """
    Get a temporary file anme with the extension name `ext`
        ext - extension name
    """
    f = tempfile.mkstemp(suffix=ext)[1]
    return f

def fastReadData(db, filename):
    """
    Use external "fast" commands to parse the input into flatAST binary data
    """
    temp_filename = get_temp_filename(".fbs")
    f2 = open(temp_filename, 'r')
    seq = f2.read()
    f2.close()
    os.system("fast -p %s %s" % (filename, temp_filename))
    f2 = open(temp_filename, 'rb')
    tree_buf = bytearray(f2.read())
    tree = Data.GetRootAsData(tree_buf, 0)
    f2.close()
    os.system("fast -S -G %s %s" % (filename, temp_filename))
    f2 = open(temp_filename, 'rb')
    graph_buf = bytearray(f2.read())
    graph = Data.GetRootAsData(graph_buf, 0)
    f2.close()
    os.remove(temp_filename)
    return [seq, tree_buf, tree, graph_buf, graph]

def readRawData(db, filename):
    """ the data is not in the database yet, read it into sequence, tree, or graph """
    seq = None
    tree_buf = None
    tree = None
    graph_buf = None
    graph = None
    _, input_extension = os.path.splitext(filename)
    if input_extension == ".fbs":      
        f = open(filename, 'rb')
        buf = f.read()
        f.close()
        data = Data.GetRootAsData(buf, 0)
        if not data is None and not data.RecordType() is None:
            if not data.RecordType().Graph() is None:
                graph_buf = buf
                graph = data
            if not data.RecordType().Element() is None:
                tree_buf = buf
                tree = data
                temp_filename = get_temp_filename(".java")
                os.system("fast -C %s > %s" % (filename, temp_filename))
                f2 = open(temp_filename, 'r')
                seq = f2.read()
                f2.close()
            return [seq, tree_buf, tree, graph_buf, graph]
    return fastReadData(db, filename)

def found_id(db, id, hash):
    conn = sqlite3.connect(db)
    c = conn.cursor()
    c = c.execute("SELECT id FROM units WHERE id = ? OR hash = ? ", (id, hash))
    rows = c.fetchone()
    conn.close()
    return not rows is None

def readData(db, filename, id=""):
    """
    Read file as GGNN data, and save into flatAST with the hash of the original file, 
    if it is not yet in the database. On the other hand, if the file is in the database, 
    and its hash is different from the stored hash, update the record with the new data.
    """
    # print("++++++++++ %s"  % filename)
    if not os.path.exists(filename):
        print("[WARN] The file %s doesn't exist." % filename)
        return None

    if id == "":
        id = filename
        
    file = open(id, 'rb')
    seq = file.read()
    hash = hashlib.sha256(seq).hexdigest()
    file.close()

    tree_buf = None
    tree = None
    graph_buf = None
    graph = None
    if not found_id(db, id, hash):
        [_, tree_buf, tree, graph_buf, graph] = readRawData(db, filename)
        """ now the data should be stored into the database """
        if not graph is None and not tree is None:
            if not found_id(db, id, hash):
                print("INSERT INTO units (id, seq, tree, graph, hash)")
                conn = sqlite3.connect(db)                
                conn.execute("INSERT INTO units (id, seq, tree, graph, hash) VALUES (?, ?, ?, ?, ?)", (id, seq, sqlite3.Binary(tree_buf), sqlite3.Binary(graph_buf), hash) )
                conn.commit()
                conn.close()
        if not graph is None:
            if not found_id(db, id, hash):
                print("INSERT INTO units (id, seq, graph, hash)")
                conn = sqlite3.connect(db)
                conn.execute("INSERT INTO units (id, seq, graph, hash) VALUES (?, ?, ?, ?)", (id, seq, sqlite3.Binary(graph_buf), hash) )
                conn.commit()
                conn.close()
        if not tree is None:
            if not found_id(db, id, hash):
                print("INSERT INTO units (id, seq, tree, hash)")                
                conn = sqlite3.connect(db)
                conn.execute("INSERT INTO units (id, seq, tree, hash) VALUES (?, ?, ?, ?)", (id, seq, sqlite3.Binary(tree_buf), hash) )
                conn.commit()
                conn.close()
    else:
        conn = sqlite3.connect(db)
        c = conn.cursor()
        c = c.execute("SELECT id, hash FROM units WHERE id = ? OR hash = ?", (id, hash))
        rows = c.fetchone()
        c.close()
        conn.close()
        if not rows is None and id == rows[0] and hash != rows[1]:
            """ update the data record """
            [seq, tree_buf, tree, graph_buf, graph] = readRawData(db, filename)
            if not graph is None:
                print("UPDATE units SET graph")
                conn = sqlite3.connect(db)
                conn.execute("UPDATE units SET graph = ?, hash = ? WHERE id = ?", 
                    (sqlite3.Binary(graph_buf), hash, id))
                conn.commit()
                conn.close()
            if not tree is None:
                print("UPDATE units SET tree")
                conn = sqlite3.connect(db)
                conn.execute("UPDATE units SET tree = ?, hash = ? WHERE id = ?", 
                    (sqlite3.Binary(tree_buf), hash, id))
                conn.commit()
                conn.close()
            if not seq is None:
                print("UPDATE units SET seq")
                conn = sqlite3.connect(db)
                conn.execute("UPDATE units SET seq = ?, hash = ? WHERE id = ?", 
                    (seq, hash, id))
                conn.commit()
                conn.close()
        elif not rows is None and id != rows[0] and hash == rows[1]:
            id = rows[0]
    return [id, seq, tree_buf, tree, graph_buf, graph]

def add_to_oracle(db, id, label=None):
    conn = sqlite3.connect(db)
    c = conn.cursor()
    c = c.execute("SELECT label FROM classes WHERE unit_id = ?", (id, ))
    rows = c.fetchone()
    c.close()
    conn.close()

    if label is None:
        """ default label is computed from the path """
        label = os.path.basename(os.path.dirname(id))
        #if label == "":
        #    print("TO TEST => %s " % label)
    if not rows is None:
        old_label = rows[0]
        if old_label != label:
            conn = sqlite3.connect(db)
            conn.execute("UPDATE classes SET label = ? WHERE unit_id = ?", (label, id) )
            conn.commit()
        return label
    conn = sqlite3.connect(db)
    conn.execute("INSERT INTO classes (unit_id, label) VALUES (?, ?)", (id, label) )
    conn.commit()
    conn.close()
    return label

def prepare_datasets_for_learning(db, splitting = True):
    """
    For each class, split instances into three sets of 6/2/2 per 10 elements
    and mark them as in the modes of training, testing and validating respectively.
    """
    """ Clean up the learning instances table """
    conn = sqlite3.connect(db)
    conn.execute("DELETE FROM learning")
    conn.commit()
    """ Select the classes """
    c = conn.cursor()
    c = c.execute("SELECT label, count(unit_id) FROM classes GROUP BY label")
    classes = c.fetchall()
    c.close()
    conn.close()
    conn = sqlite3.connect(db)
    units = []
    class_names = []
    for i in range(0, len(classes)):
        """ randomly split instances in the class by 6/2/2 or 7/3/0 ratio into training, 
        testing, validating mode """
        n = classes[i][1]
        # print("Group %s: %d" % (classes[i][0], n))
        if not classes[i][0] == "":
            class_names.append(classes[i][0])
        c = conn.cursor()
        c = c.execute("SELECT unit_id FROM classes WHERE label = ?", (classes[i][0],))
        ids = c.fetchall()
        c.close()
        if n > 0:
            x = random.sample(range(0,n), n)
            for k in range(0, len(ids)):
                j = float(10 * x[k]) / len(ids)
                if classes[i][0] == "":
                    mode = 2       # Testing
                    units.append(ids[k][0])
                else:
                    if splitting and j < 7.0:
                        mode = 0   # Training
                    else:
                        mode = 1   # Validating
                # if j < 6.0:
                #     mode = 0   # Training
                # elif j < 8.0:
                #     mode = 1   # Validating
                # else:
                #     mode = 2   # Testing
                conn.execute("INSERT INTO learning (unit_id, mode) VALUES (?, ?)", (ids[k][0], mode))
                conn.commit()
    conn.close()
    return units, class_names

def prepare_logs_for_training(log_path):
    if log_path != "":
        previous_runs = os.listdir(log_path)
        if len(previous_runs) == 0:
            run_number = 1
        else:
            run_number = max([int(s.split("run-")[1]) for s in previous_runs]) + 1
        writer = SummaryWriter("%s/run-%03d" % (log_path, run_number))
    else:
        writer = None
    return writer

def load_model_at_epoch(opt, opt_model_path, opt_epoch):
    filename = "{}.{}".format(opt_model_path, opt_epoch)
    epoch = opt_epoch
    if os.path.exists(filename):
        dirname = os.path.dirname(filename)
        basename = os.path.basename(filename)
        epochs = os.listdir(dirname)
        if len(epochs) > 0:
            for s in epochs:
                if s.startswith(basename) and basename != s:
                    x = s.split(os.extsep)
                    e = x[len(x) - 1]
                    epoch = max(epoch, int(e))
        if epoch != -1:
            # print("Using No. {} of the saved models...".format(epoch))
            filename = "{}.{}".format(opt_model_path, epoch)
        net = torch.load(filename)
    else:
        net = GGNN(opt)
        net.double()
    return epoch, net

def delete_validation_data(db):
    conn = sqlite3.connect(db)
    conn.execute("DELETE FROM units WHERE id IN (SELECT unit_id FROM classes WHERE label='') ")
    conn.execute("DELETE FROM classes WHERE label=''")
    conn.commit()
    conn.close()

def main():
    """
    We store data into the three tables:
    -  units (id, seq, tree, graph, hash) index the data to be classified by a unique id, where data
            can be any structure in terms of sequence (seq), abstract syntax tree (tree), and/or gated graphs (graph). 
            In our case, these structures follow the flatAST schema. 
            The hash is sha256sum of the data's original content. In case the file is updated 
            with a hash different from the stored hash, we will update the record. 
    -  classes (unit_id, label) is the oracle of classifications of these units
    -  learning (unit_id, mode) is the data used for training, testing, and validation,
            during the machine learning process.
    """

    try:
        start = time.time()
        conn = sqlite3.connect(opt.db)
        conn.execute('''CREATE TABLE IF NOT EXISTS units(id TEXT PRIMARY KEY, seq TEXT, tree BLOB, graph BLOB, hash TEXT)''')
        conn.execute('''CREATE TABLE IF NOT EXISTS classes(unit_id TEXT PRIMARY KEY, label TEXT)''')
        conn.execute('''CREATE TABLE IF NOT EXISTS learning(unit_id TEXT PRIMARY KEY, mode INTEGER)''')
        conn.close()

        delete_validation_data(opt.db)
        for i in range(0, len(opt.argv)):  # adding data 
            # print("adding the file %s ..." % opt.argv[i])
            [id, seq, tree_buf, tree, graph_buf, graph] = readData(opt.db, opt.argv[i])
            print(id)
            if id != opt.argv[i]:
                print("The file %s is already in the dataset as %s" % (opt.argv[i], id))
            else:
                add_to_oracle(opt.db, id, opt.label)
        torch.manual_seed(opt.manualSeed)
        if opt.cuda:
            torch.cuda.manual_seed_all(opt.manualSeed)

        units, classes = prepare_datasets_for_learning(opt.db)

        if len(opt.argv) == 0:
            train_dataset = ProgramData(opt, opt.map_folder, opt.size_vocabulary, opt.db, 0, opt.n_classes)
            train_dataloader = bAbIDataloader(train_dataset, batch_size=opt.train_batch_size, shuffle=True, num_workers=2)

            valid_dataset = ProgramData(opt, opt.map_folder, opt.size_vocabulary, opt.db, 1, opt.n_classes)
            valid_dataloader = bAbIDataloader(valid_dataset, batch_size=opt.test_batch_size, shuffle=True, num_workers=2)
            opt.n_edge_types = train_dataset.n_edge_types
            opt.n_node = train_dataset.n_node

        else:
            test_dataset = ProgramData(opt, opt.map_folder, opt.size_vocabulary, opt.db, 2, opt.n_classes)
            if len(test_dataset) > 0:
                test_dataloader = bAbIDataloader(test_dataset, batch_size=opt.test_batch_size, shuffle=True, num_workers=2)
            opt.n_edge_types = test_dataset.n_edge_types
            opt.n_node = test_dataset.n_node

        opt.annotation_dim = 1  # for bAbI
        epoch, net = load_model_at_epoch(opt, opt.model_path, opt.epoch)
        criterion = nn.CrossEntropyLoss()
        if opt.cuda:
            net.cuda()
            criterion.cuda()
        optimizer = optim.Adam(net.parameters(), lr=opt.lr)

        if len(opt.argv) == 0:
            # print("Training and validating ...")
            writer = prepare_logs_for_training(opt.log_path)
            for epoch in range(epoch+1, epoch + opt.niter):
                train(epoch, train_dataloader, net, criterion, optimizer, opt, writer)
                if not opt.is_training_ggnn:
                    test(valid_dataloader, net, criterion, optimizer, opt, units, classes, True)
            test(valid_dataloader, net, criterion, optimizer, opt, units, classes, True)
            if not writer is None:
                writer.close()
        else:
            if len(test_dataset) > 0:
                # print("Testing...")
                test(test_dataloader, net, criterion, optimizer, opt, units, classes, False)
        end = time.time()
        print(end - start)
        os._exit(0)
    except:
        typ, value, tb = sys.exc_info()
        traceback.print_exc()
        pdb.post_mortem(tb)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import argparse
import os
import os.path
import pickle
from collections import OrderedDict

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--lang', default='cpp', help='language')
    opt = parser.parse_args()
    print(opt)
    maps = {}
    map_filename = 'maps.%s.pkl' % opt.lang
    if os.path.exists(map_filename):
       with open(map_filename, 'rb') as f:
            maps = pickle.load(f)
    OrderedDict(sorted(maps.items(), key=lambda t: t[0]))
    for key, value in maps.items() :
        print("%s<=%s" % (value, key))

if __name__ == "__main__":
    main()




import torch
from torch.autograd import Variable
from tensorboardX import SummaryWriter
from shutil import copyfile

def train(epoch, dataloader, net, criterion, optimizer, opt, writer):
    
    for i, (adj_matrices, target) in enumerate(dataloader, 0):
    
        net.zero_grad()
        # optimizer.zero_grad()
        left_adj_matrix = adj_matrices[0]
        right_adj_matrix = adj_matrices[1]
    
        left_init_input = torch.zeros(len(left_adj_matrix), opt.n_node, opt.state_dim).double()
        right_init_input = torch.zeros(len(right_adj_matrix), opt.n_node, opt.state_dim).double()

        if opt.cuda:
            # print("Using cuda for training.......")
            left_init_input = left_init_input.cuda()
            right_init_input = right_init_input.cuda()
            left_adj_matrix = left_adj_matrix.cuda()
            right_adj_matrix = right_adj_matrix.cuda()
            target = target.cuda()

        left_init_input = Variable(left_init_input)
        right_init_input = Variable(right_init_input)

        left_adj_matrix = Variable(left_adj_matrix)
        right_adj_matrix = Variable(right_adj_matrix)

        target = Variable(target)
        # print(target)

        if opt.loss == 1:
            left_output, right_output = net(left_init_input, left_adj_matrix, right_init_input, right_adj_matrix)
            loss = criterion(left_output,right_output, target) 
            if writer:
               writer.add_scalar('loss', loss.data.item(), int(epoch))
        else:
            output = net(left_init_input, left_adj_matrix, right_init_input, right_adj_matrix)
            loss = criterion(output, target) 
            if writer:
               writer.add_scalar('loss', loss.data.item(), int(epoch))
           
        loss.backward()
        optimizer.step()

        if i % int(len(dataloader) / 10 + 1) == 0 and opt.verbal:
            print('[%d/%d][%d/%d] Loss: %.4f' % (epoch, opt.niter, i, len(dataloader), loss.item()))

    torch.save(net, opt.model_path)
    copyfile(opt.model_path, "{}.{}".format(opt.model_path, epoch))



import torch
from torch.autograd import Variable
from tensorboardX import SummaryWriter
from shutil import copyfile
import os
import sys

def train(epoch, dataloader, net, criterion, optimizer, opt, writer):
    
    for i, (adj_matrix, target) in enumerate(dataloader, 0):
        net.zero_grad()

        # padding = torch.zeros(len(annotation), opt.n_node, opt.state_dim - opt.annotation_dim).double()
        # init_input = torch.cat((annotation, padding), 2)
        init_input = torch.zeros(len(adj_matrix), opt.n_node, opt.state_dim).double()

        if opt.cuda:
            init_input = init_input.cuda()
            adj_matrix = adj_matrix.cuda()
            # annotation = annotation.cuda()
            target = target.cuda()

        init_input = Variable(init_input)
        adj_matrix = Variable(adj_matrix)
        # annotation = Variable(annotation)
        target = Variable(target)
        output = net(init_input, adj_matrix)
        loss = criterion(output, target)
        loss.backward()
        optimizer.step()

        if not writer is None:
            writer.add_scalar('loss', loss.data.item(), int(epoch))
        if i % int(len(dataloader) / 10 + 1) == 0 and opt.verbal:
            print('[%d/%d][%d/%d] Loss: %.4f' % (epoch, opt.niter, i, len(dataloader), loss.item()))

    if (not os.path.exists(os.path.dirname(opt.model_path))):
        os.mkdir(os.path.dirname(opt.model_path))
    torch.save(net, opt.model_path)
    copyfile(opt.model_path, "{}.{}".format(opt.model_path, epoch))



import numpy as np
import os
from os import listdir
from os.path import isfile, join
import collections
import re
from tqdm import trange
import random
import pickle
import pyarrow
import sqlite3
import fast_
import fast_.Graph_
import fast_.Data_
import fast_.Data_.Anonymous3
from fast_ import Data, Graph, Element, Bugs, Log, Delta
from fast_.Data_ import Anonymous3
from fast_.Data import *
from fast_.Graph import *
from fast_.Element import *
from fast_.Bugs import *
from fast_.Log import *
from fast_.Delta import *
from fast_.Data_.Anonymous3 import *
from fast_.Graph import *
from fast_.Graph_.Unit import *
from fast_.Graph_.ContextGraph import *
from fast_.Graph_.ContextEdges import *
from fast_.Graph_.Edge import *
from fast_.Graph_.SymbolCandidate import *
from fast_.Graph_.NodeLabel import *
from fast_.Graph_.NodeType import *
import flatast.json_format
import flatast.fast_pb2
import utils

def load_graphs_from_file(file_name):
    data_list = []
    edge_list = []
    target_list = []    
    with open(file_name,'r') as f:
        for line in f:
            if len(line.strip()) == 0:
                data_list.append([edge_list,target_list])
                edge_list = []
                target_list = []
            else:
                digits = []
                line_tokens = line.split(" ")

                if line_tokens[0] == "?":
                    for i in range(1, len(line_tokens)):
                        
                        digits.append(int(line_tokens[i]))
                    target_list.append(digits)
                else:
                    for i in range(len(line_tokens)):
                        digits.append(int(line_tokens[i]))
                    edge_list.append(digits)
    return data_list

def load_program_graphs_from_directory(directory,is_train=True,n_classes=3, data_percentage=1.0):
    data_list = []
    if is_train == True:
            dir_path =  os.path.join(directory,"train")
    else:
            dir_path =  os.path.join(directory,"test")
    filenames = []
    for f in listdir(dir_path):
        if isfile(join(dir_path, f)):
            filenames.append(f)
    int_filenames = [int(re.search('_(.*).txt', x).group(1)) for x in filenames]
    ordered_filenames = sorted(int_filenames)
    lookup = {}
    for i in range(1, 1+len(ordered_filenames)):
        if is_train == True:
            lookup[i] = join(dir_path, "train_%s.txt" % str(ordered_filenames[i-1]))
        else:
            lookup[i] = join(dir_path, "test_%s.txt" % str(ordered_filenames[i-1]))
    for i in trange(1, 1+n_classes):
        path = lookup[i]
        print(path)
        label = i
        data_list_class_i = []
        edge_list_class_i = []
        target_list_class_i = []

        with open(path,'r') as f:
            for line in f: 
                if len(line.strip()) == 0:

                    data_list_class_i.append([edge_list_class_i,target_list_class_i])
                    edge_list_class_i = []
                    target_list_class_i = []
                else:
                    digits = []
                    line_tokens = line.split(" ")
                    
                    if line_tokens[0] == "?":

                        target_list_class_i.append([label])
                    else:
                        for j in range(len(line_tokens)):
                            digits.append(int(line_tokens[j]))
                        edge_list_class_i.append(digits)

        if data_percentage < 1.0:
            print("Cutting down " + str(data_percentage) + " of all data......")
            slicing = int(len(data_list_class_i)*data_percentage)
            print("Remaining data : " + str(slicing) + "......")
            data_list_class_i = data_list_class_i[:slicing]

        data_list.extend(data_list_class_i)

    return data_list

def load_program_graphs_from_db(opt, db, map_folder=".", mode=0, n_classes=3):
    """
    Load the program graph data from the database db
    """       
    """ Select the classes """
    conn = sqlite3.connect(db)
    c = conn.cursor()
    c = c.execute("SELECT c.label, count(c.unit_id) FROM classes c, learning l WHERE c.unit_id = l.unit_id AND l.mode = ? GROUP BY c.label", (mode,))
    classes = c.fetchall()
    c.close()
    conn.close()
    conn = sqlite3.connect(db)
    data_list = []
    algorithms = []
    maps = {} 
    maps_filename = None
    # print("======> %d" % len(classes))
    for i in range(0, len(classes)):
        label = classes[i][0]
        n_instance = classes[i][1]
        if n_instance == 0:
            continue
        conn = sqlite3.connect(db)
        c = conn.cursor()
        c = c.execute("SELECT u.graph, u.id, c.label, l.mode FROM units u, classes c, learning l WHERE u.id = c.unit_id AND c.unit_id = l.unit_id AND l.mode = ? AND c.label = ?", (mode, label))
        rows = c.fetchall()
        c.close()
        conn.close()
        n_instance = len(rows)
        edge_list_class_i = []
        target_list_class_i = []
        data_list_class_i = []
        # print("======> %d" % n_instance)
        for j in range(0, n_instance):
            buf = rows[j][0]
            if rows[j][3] == 2:
                print("%s: to test" % rows[j][1])
            elif rows[j][3] == 0:
                print("%s -> %s: to train" % (rows[j][1], rows[j][2]))
            else:
                print("%s -> %s: to validate" % (rows[j][1], rows[j][2]))
            data = Data.GetRootAsData(buf, 0)
            graph = data.RecordType().Graph()
            if not graph is None:  # in case the graph has been correctly rendered
                for k in range(0, graph.UnitLength()):
                    uniq_edges = set()
                    unit = graph.Unit(k)
                    from ggnn import ggnn
                    (maps_filename, uniq_edges) = ggnn.ggnn2str(opt, unit, uniq_edges, map_folder, algorithms, maps)
                    for e in set(uniq_edges):
                        digits = e.split()
                        edge = []
                        for l in range(0, len(digits)):
                            edge.append(int(digits[l]))
                        edge_list_class_i.append(edge)
                target_list_class_i.append([i])
                data_list_class_i.append([edge_list_class_i,target_list_class_i])
        data_list.extend(data_list_class_i)
    if (opt.maps or opt.edgemaps) and not opt.localmaps and not maps_filename is None:
        # Don't assume the files in the same dataset are of the same language
        with open(maps_filename, 'wb') as f:
            pickle.dump(maps, f, 2)
    # print("======> %d" % len(data_list))
    return data_list

def find_max_edge_id(data_list):
    max_edge_id = 0
    for data in data_list:
        edges = data[0]
        for item in edges:
            if item[1] > max_edge_id:
                max_edge_id = item[1]
    return max_edge_id

def find_max_node_id(data_list):
    max_node_id = 0
    for data in data_list:
        edges = data[0]
        for item in edges:
            if item[0] > max_node_id:
                max_node_id = item[0]
            if item[2] > max_node_id:
                max_node_id = item[2]
    return max_node_id

def find_max_task_id(data_list):
    max_node_id = 0
    for data in data_list:
        target = data[1]
        for item in target:
            if item[0] > max_node_id:
                max_node_id = item[0]
    return max_node_id

def convert_program_data(data_list, n_annotation_dim, n_nodes):
    # n_nodes = find_max_node_id(data_list)
    class_data_list = []

    for item in data_list:
        edge_list = item[0]
        target_list = item[1]
        for target in target_list:
            task_type = target[0]
            task_output = target[-1]
            annotation = np.zeros([n_nodes, n_annotation_dim])   
            for edge in edge_list:
                src_idx = edge[0]
                if src_idx < len(annotation):
                    annotation[src_idx-1][0] = 1        
        class_data_list.append([edge_list, annotation, task_output])
    return class_data_list

def convert_program_data_into_group(data_list, n_annotation_dim, n_nodes, n_classes):
    class_data_list = []

    for i in range(n_classes):
        class_data_list.append([])

    for item in data_list:
        edge_list = item[0]
        target_list = item[1]
        for target in target_list:
            class_output = target[-1]
            annotation = np.zeros([n_nodes, n_annotation_dim])   
            for edge in edge_list:
                src_idx = int(edge[0])
                if src_idx < len(annotation):
                    annotation[src_idx-1][0] = 1
            class_data_list[class_output-1].append([edge_list, annotation, class_output])
    return class_data_list

def create_adjacency_matrix(edges, n_nodes, n_edge_types):
    a = np.zeros([n_nodes, n_nodes * n_edge_types * 2])

    for edge in edges:
        src_idx = edge[0]
        e_type = edge[1]
        tgt_idx = edge[2]
    
        if tgt_idx < len(a):
            a[tgt_idx-1][(e_type - 1) * n_nodes + src_idx - 1] =  1
        if src_idx <len(a):
            a[src_idx-1][(e_type - 1 + n_edge_types) * n_nodes + tgt_idx - 1] =  1
    return a

class MonoLanguageProgramData():   
    def __init__(self, size_vocabulary, path, is_train, n_classes=3,data_percentage=1.0):
        base_name = os.path.basename(path)
        if is_train:
            saved_input_filename = "%s/%s-%d-train.pkl" % (path, base_name, n_classes)
        else:
            saved_input_filename = "%s/%s-%d-test.pkl" % (path, base_name, n_classes)
        if os.path.exists(saved_input_filename): 
            input_file = open(saved_input_filename, 'rb')
            buf = input_file.read()
            all_data = pyarrow.deserialize(buf)
            input_file.close()
        else:
            all_data = load_program_graphs_from_directory(path,is_train,n_classes,data_percentage)
            all_data = np.array(all_data)[0:len(all_data)]
            buf = pyarrow.serialize(all_data).to_buffer()
            out = pyarrow.OSFile(saved_input_filename, 'wb')
            out.write(buf)
            out.close()
    
        if is_train == True:
            print("Number of all training data : " + str(len(all_data)))
        else:
            print("Number of all testing data : " + str(len(all_data)))
        self.n_edge_types =  find_max_edge_id(all_data)
        # print("Edge types : " + str(self.n_edge_types))
        max_node = find_max_node_id(all_data)
        print("Max node id : " + str(max_node))
        self.n_node = size_vocabulary
        
        all_data = convert_program_data(all_data,1, self.n_node)

        self.data = all_data
    
    def __getitem__(self, index):
        
        am = create_adjacency_matrix(self.data[index][0], self.n_node, self.n_edge_types)
        # annotation = self.data[index][1]
        target = self.data[index][2] - 1
        return am, target

    def __len__(self):
        return len(self.data)

class ProgramData():   
    def __init__(self, opt, map_folder, size_vocabulary, db, mode, n_classes=3):
        all_data = load_program_graphs_from_db(opt, db, map_folder, mode, n_classes)
        all_data = np.array(all_data)[0:len(all_data)]
        if False:
            if mode == 0:
                print("Number of all training data : " + str(len(all_data)))
            elif mode == 1:
                print("Number of all validating data : " + str(len(all_data)))
            else:
                print("Number of all testing data : " + str(len(all_data)))
        self.n_edge_types =  find_max_edge_id(all_data)
        max_node = find_max_node_id(all_data)
        # print("Max node id : " + str(max_node))
        self.n_node = size_vocabulary        
        # print("======> %d" % len(all_data))
        all_data = convert_program_data(all_data, 1, self.n_node)
        # print("======> %d" % len(all_data))
        self.data = all_data
    
    def __getitem__(self, index):        
        am = create_adjacency_matrix(self.data[index][0], self.n_node, self.n_edge_types)
        # annotation = self.data[index][1]
        target = self.data[index][2]
        return am, target

    def __len__(self):
        return len(self.data)

class CrossLingualProgramData():

    def __init__(self, size_vocabulary, left_path, right_path, is_train, loss, n_classes=3, data_percentage=1):
        
        self.loss = loss
        left_all_data = load_program_graphs_from_directory(left_path,is_train,n_classes,data_percentage)
        right_all_data = load_program_graphs_from_directory(right_path,is_train,n_classes,data_percentage)

        left_all_data = np.array(left_all_data)[0:len(left_all_data)]
        right_all_data = np.array(right_all_data)[0:len(right_all_data)]
    
        if is_train == True:
            print("Number of all left training data : " + str(len(left_all_data)))
            print("Number of all right training data : " + str(len(right_all_data)))
        else:
            print("Number of all left testing data : " + str(len(left_all_data)))
            print("Number of all right testing data : " + str(len(right_all_data)))

        self.n_edge_types =  find_max_edge_id(left_all_data)
        self.n_node = size_vocabulary
        max_left_node = find_max_node_id(left_all_data)
        max_right_node = find_max_node_id(right_all_data)

        print("Left max node id : " + str(max_left_node))
        print("Right max node id : " + str(max_right_node))

        left_all_data_by_classes = convert_program_data_into_group(left_all_data,1, self.n_node, n_classes)

        right_all_data_by_classes = convert_program_data_into_group(right_all_data,1, self.n_node, n_classes)

        pairs_1 = []
        pairs_0 = []

        for i, left_class in tqdm(enumerate(left_all_data_by_classes)):
            right_class = right_all_data_by_classes[i]

            remaining_right_class = []

            for j, other_right_class in enumerate(right_all_data_by_classes):
                if j != i:
                    remaining_right_class.extend(other_right_class)

            if len(left_class) > len(right_class):
                left_class = left_class[:len(right_class)]
            
            for k, left_data_point in enumerate(left_class):
                righ_data_point = right_class[k]
                pairs_1.append((left_data_point,righ_data_point))
                pairs_0.append((left_data_point, random.choice(remaining_right_class)))

        print("Number of all 1 pairs data : " + str(len(pairs_1)))
        print("Number of all 0 pairs data : " + str(len(pairs_0)))
        data = []
        data.extend(pairs_1)
        data.extend(pairs_0)
        random.shuffle(data)
        self.data = data

    
    def __getitem__(self, index):
        
        left_data_point = self.data[index][0]
        right_data_point = self.data[index][1]

        left_am = create_adjacency_matrix(left_data_point[0], self.n_node, self.n_edge_types)
        right_am = create_adjacency_matrix(right_data_point[0], self.n_node, self.n_edge_types)

        left_annotation = left_data_point[1]
        right_annotation = right_data_point[1]

        if left_data_point[2] == right_data_point[2]:
            target = 1.0
        else:
            target = 0.0

        if self.loss == 0:
            target = int(target)

        return (left_am,right_am), target

    def __len__(self):
        return len(self.data)




import torch
from torch.autograd import Variable
from sklearn.metrics import classification_report, confusion_matrix, accuracy_score

def test(dataloader, net, criterion, optimizer, opt, units, classes, accuracy = True):
    test_loss = 0
    correct = 0
    net.eval()

    all_targets = []
    all_predicted = []

    for i, (adj_matrix, target) in enumerate(dataloader, 0):
        
        # padding = torch.zeros(len(annotation), opt.n_node, opt.state_dim - opt.annotation_dim).double()
        # init_input = torch.cat((annotation, padding), 2)
        init_input = torch.zeros(len(adj_matrix), opt.n_node, opt.state_dim).double()
        if opt.cuda:
            init_input = init_input.cuda()
            adj_matrix = adj_matrix.cuda()
            # annotation = annotation.cuda()
            target = target.cuda()

        init_input = Variable(init_input)
        adj_matrix = Variable(adj_matrix)
        # annotation = Variable(annotation)
        target = Variable(target)
        # print(target)
        output = net(init_input, adj_matrix)
        # print(output)
        #test_loss += criterion(output, target).data[0]
        test_loss += criterion(output, target).item()
        pred = output.data.max(1, keepdim=True)[1]
        if not accuracy:
            print(pred)
            for u in range(0, len(units)):
                print("%s is %s" % (units[u], classes[pred[u][0]]))
        all_predicted.extend(pred.data.view_as(target).cpu().numpy())
        all_targets.extend(target.cpu().numpy())
        correct += pred.eq(target.data.view_as(pred)).cpu().sum()

    test_loss /= len(dataloader.dataset)
    if accuracy:
        print('Accuracy:', accuracy_score(all_targets, all_predicted))
        print(classification_report(all_targets, all_predicted))
        print(confusion_matrix(all_targets, all_predicted))
        print('Test set: Average loss: {:.4f}, Accuracy: {}/{} ({:.0f}%)'.format(
            test_loss, correct, len(dataloader.dataset),
            100. * correct / len(dataloader.dataset)))



#!/usr/bin/env python3
import argparse
import os
import os.path
import pickle
from collections import OrderedDict

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--lang1', default='cpp', help='source language')
    parser.add_argument('--lang2', default='swap', help='swapped language')
    parser.add_argument('id', nargs='+', help='id to swap')
    opt = parser.parse_args()
    print(opt)
    maps = {}
    maps1 = {}
    maps1_filename = 'maps.%s.pkl' % opt.lang1
    if os.path.exists(maps1_filename):
       with open(maps1_filename, 'rb') as f:
            maps1 = pickle.load(f)
    OrderedDict(sorted(maps1.items(), key=lambda t: t[0]))

    maps2 = {}
    for (k,v) in maps1.items():
            maps2[k] = v
    n = len(opt.id)
    for i in range(0, n-1):
            i1 = int(opt.id[i])
            for (k, v) in maps1.items():
                    if int(v) == i1:
                            maps2[k] = opt.id[i+1]
    i1 = int(opt.id[n-1])
    for (k, v) in maps1.items():
            if int(v) == i1:
                    maps2[k] = opt.id[0]

    maps2_filename = 'maps.%s.pkl' % opt.lang2
    with open(maps2_filename, 'wb') as f:
            pickle.dump(maps2, f)

if __name__ == "__main__":
    main()



import torch
import torch.nn as nn
import torch.nn.init as init
import torch.nn.functional as F
from tensorboardX import SummaryWriter

class AttrProxy(object):
    """
    Translates index lookups into attribute lookups.
    To implement some trick which able to use list of nn.Module in a nn.Module
    see https://discuss.pytorch.org/t/list-of-nn-module-in-a-nn-module/219/2
    """
    def __init__(self, module, prefix):
        self.module = module
        self.prefix = prefix

    def __getitem__(self, i):
        return getattr(self.module, self.prefix + str(i))


class Propogator(nn.Module):
    """
    Gated Propogator for GGNN
    Using LSTM gating mechanism
    """
    def __init__(self, state_dim, n_node, n_edge_types):
        super(Propogator, self).__init__()

        self.n_node = n_node
        self.n_edge_types = n_edge_types

        self.reset_gate = nn.Sequential(
            nn.Linear(state_dim*3, state_dim),
            nn.Sigmoid()
        )
        self.update_gate = nn.Sequential(
            nn.Linear(state_dim*3, state_dim),
            nn.Sigmoid()
        )
        self.tansform = nn.Sequential(
            nn.Linear(state_dim*3, state_dim),
            nn.Tanh()
        )

    def forward(self, state_in, state_out, state_cur, A):
        A_in = A[:, :, :self.n_node*self.n_edge_types]
        A_out = A[:, :, self.n_node*self.n_edge_types:]

        a_in = torch.bmm(A_in, state_in)
        a_out = torch.bmm(A_out, state_out)
        a = torch.cat((a_in, a_out, state_cur), 2)

        r = self.reset_gate(a)
        z = self.update_gate(a)
        
        joined_input = torch.cat((a_in, a_out, r * state_cur), 2)
        h_hat = self.tansform(joined_input)

        output = (1 - z) * state_cur + z * h_hat

        return output

class ClassPrediction(nn.Module):

    def __init__(self, opt):
        super(ClassPrediction, self).__init__()

        self.class_prediction = nn.Sequential(
            nn.Linear(opt.n_node, opt.n_hidden),
            nn.Tanh(),
            nn.Linear(opt.n_hidden, opt.n_classes),
            nn.Softmax(dim=1)    
        )

        self.criterion = nn.CrossEntropyLoss()
        self._initialization()

    def _initialization(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                init.xavier_normal_(m.weight.data)
                if m.bias is not None:
                    init.normal_(m.bias.data)

    def forward(self, graph_representation, target):
        output = self.class_prediction(graph_representation)
        loss = self.criterion(output, target)
       
        return loss

class ContrastiveLoss(nn.Module):
    def __init__(self, margin=0.5):
        super(ContrastiveLoss, self).__init__()
        self.margin = margin

    def forward(self, output1, output2, label):
        euclidean_distance = F.pairwise_distance(output1, output2)
        loss_contrastive = torch.mean((1.0 - label) * torch.pow(euclidean_distance, 2) + (label) * torch.pow(torch.clamp(self.margin - euclidean_distance, min=0.0), 2))
        return loss_contrastive

class GGNN(nn.Module):
    """
    Gated Graph Sequence Neural Networks (GGNN)
    Mode: SelectNode
    Implementation based on https://arxiv.org/abs/1511.05493
    """
    def __init__(self, opt):
        super(GGNN, self).__init__()

        # assert (opt.state_dim >= opt.annotation_dim, 'state_dim must be no less than annotation_dim')
        self.is_training_ggnn = opt.is_training_ggnn
        self.state_dim = opt.state_dim
        self.n_edge_types = opt.n_edge_types
        self.n_node = opt.n_node
        self.n_steps = opt.n_steps
        self.n_classes = opt.n_classes

        for i in range(self.n_edge_types):
            # incoming and outgoing edge embedding
            in_fc = nn.Linear(self.state_dim, self.state_dim)
            out_fc = nn.Linear(self.state_dim, self.state_dim)
            self.add_module("in_{}".format(i), in_fc)
            self.add_module("out_{}".format(i), out_fc)

        self.in_fcs = AttrProxy(self, "in_")
        self.out_fcs = AttrProxy(self, "out_")

        # Propogation Model
        self.propogator = Propogator(self.state_dim, self.n_node, self.n_edge_types)

        # Output Model
        self.out = nn.Sequential(
            nn.Linear(self.state_dim, self.state_dim),
            nn.LeakyReLU(),
            nn.Linear(self.state_dim, 1),
            nn.Tanh(),   
        )

      
        self.soft_attention = nn.Sequential(
            nn.Linear(self.state_dim, self.state_dim),
            nn.LeakyReLU(),
            nn.Linear(self.state_dim, 1),
            nn.Sigmoid(),
           
        )

        self.class_prediction = nn.Sequential(
            nn.Linear(opt.n_node, opt.n_hidden),
            nn.Tanh(),
            nn.Linear(opt.n_hidden, opt.n_classes),
            nn.Softmax(dim=1)    
        )

        # self.class_prediction = nn.Sequential(
        #     nn.Linear(opt.n_node, opt.n_classes),
        #     nn.Softmax(dim=1)    
        # )

        self._initialization()

    def _initialization(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                init.xavier_normal_(m.weight.data)
                if m.bias is not None:
                    init.normal_(m.bias.data)

    def forward(self, prop_state, A):
        # print(prop_state.shape)
        for i_step in range(self.n_steps):
            in_states = []
            out_states = []
            for i in range(self.n_edge_types):
                in_states.append(self.in_fcs[i](prop_state))
                out_states.append(self.out_fcs[i](prop_state))
    
            in_states = torch.stack(in_states).transpose(0, 1).contiguous()
            in_states = in_states.view(-1, self.n_node*self.n_edge_types, self.state_dim)

            out_states = torch.stack(out_states).transpose(0, 1).contiguous()
            out_states = out_states.view(-1, self.n_node*self.n_edge_types, self.state_dim)

            prop_state = self.propogator(in_states, out_states, prop_state, A)
       
        # print("Prop state : " + str(prop_state.shape))
        output = self.out(prop_state)

        # print("Out : " + str(output.shape))

        soft_attention_ouput = self.soft_attention(prop_state)
        # Element wise hadamard product to get the graph representation, check Equation 7 in GGNN paper for more details
        output = torch.mul(output,soft_attention_ouput)
        output = output.sum(2)

        if self.is_training_ggnn == True:
            output = self.class_prediction(output)

        return output


class BiGGNN(nn.Module):
    def __init__(self, opt):
        super(BiGGNN, self).__init__()

        self.opt = opt
        self.ggnn = GGNN(opt)
  
        self.n_node = opt.n_node

        self.fc_output = nn.Sequential(
            nn.Linear(20*2, 50),
            nn.ReLU(),
            nn.Linear(50, 2),
            nn.Softmax(dim=1)
        )

        self.feature_extraction = nn.Sequential(
            nn.Linear(self.n_node, 50),
            nn.ReLU(),
            nn.Linear(50, 20)  
        )

        self._initialization()

    def _initialization(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                init.xavier_normal_(m.weight.data)
                if m.bias is not None:
                    init.normal_(m.bias.data)

    def forward(self, left_prop_state, left_A, right_prop_state, right_A):
        
        # self.ggnn.zero_grad()
    
        left_output = self.ggnn(left_prop_state, left_A)
        right_output = self.ggnn(right_prop_state, right_A)

        left_output = self.feature_extraction(left_output)
        right_output = self.feature_extraction(right_output)

        if self.opt.loss == 1:
            return left_output, right_output
        else:
            concat_layer = torch.cat((left_output, right_output),1)
            output = self.fc_output(concat_layer)
            return output
